/****************************************************************************
** Demonstration of mapping a 64x64 bitmap onto a foursided polygon        **
**  by Steven H Don                                                        **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************/
#include <conio.h>
#include <dos.h>
#include <malloc.h>
#include <values.h>
#include <stdio.h>
#include <math.h>
#ifdef __DJGPP
  #include <sys\nearptr.h>
  #include <pc.h>
  #include <go32.h>
  #include <dpmi.h>
#endif

//A pointer to the VGA memory
#ifdef __DJGPP
  char *Screen;
#else
  char far *Screen = (char far*)MK_FP (0xA000, 0000);
#endif

//This sets the display to VGA 320x200 in 256 colours
void VGAScreen ()
{
  union REGS r;
  r.h.ah = 0;
  r.h.al = 0x13;
  int86(0x10, &r, &r);
}

//This resets the display to text mode
void TextScreen ()
{
  union REGS r;
  r.h.ah = 0;
  r.h.al = 0x3;
  int86(0x10, &r, &r);
}

//Sets a on the screen to a specific colour
void PutPixel (unsigned int x, unsigned int y, unsigned char c)
{
  if ((x < 320) && (y < 200))
    Screen [(y << 8) + (y << 6) + x] = c;
}

/*This draws one scanline of the textured polygon, it expects start- and
endpoints of a horizontal line and two sets of texture coordinates. It
interpolates between them. The coordinates must be in 6:16 bit precision.
(that means, the real coordinates, with a range of 0..63, shifted 16 bits to
the left).*/
void DrawTLine (int x1, int x2, int y,
                long u1, long v1, long u2, long v2,
                char *Sprite)
{
  long x, dx, du, dv, u, v, tu, tv;

  //Calculate the length of the line
  dx = x2 - x1;

  //Start at the first set of texture coordinates
  u = u1;
  v = v1;

  //Calculate the difference in the two sets of coordinates
  du = u2 - u1;
  dv = v2 - v1;

  /*If the length of the scan line is longer than 0, we can calculate the
  slope. We use this to interpolate between the first and second sets of
  texture coordinates*/
  if (dx > 0) {
    du = du / dx;
    dv = dv / dx;
  }

  //Then, we process every pixel on the scan line
  for (x = x1; x <= x2; x++) {
    /*As we use 6:16 bit precision, we have to compensate this to get the
    real texture coordinates*/
    tu = u >> 16;
    tv = v >> 16;

    //Display the correct texel (=textured pixel)
    PutPixel (x, y, Sprite [(tv << 6) + tu]);

    //Move on to the next set of coordinates
    u += du;
    v += dv;
  }
}

/*This is the main void of the texture mapping function. It interpolates
between the vertices/sides of the polygon. It gives us the start- and
endpoints of the scanlines and the appropriate texture coordinates.
The input consists of four vertices forming a four-sided polygon. Because
we work on a scanline basis, there may be no more than 2 intersection of a
scanline with a side of a polygon.*/
void TextureMap (long x1, long y1, long x2, long y2,
                 long x3, long y3, long x4, long y4,
                 char *Sprite)
/*Note: the sides and vertices 1, 2, 3 and 4 are arranged as follows:

          side 1
    1----------------2
    |                |
  s |  +------- u    | s
  i |  |             | i
  d |  |             | d
  e |  |             | e
    |  |             |
  4 |  v             | 2
    |                |
    4----------------3
          side 3

The texture will be mapped onto the polygon accordingly, the texture
coordinates running as shown above.*/
{
  long x, y,
  minx, miny, minu, minv,
  maxx, maxy, maxu, maxv,
  tx1, tx2, tx3, tx4, dx1, dx2, dx3, dx4,
  dy1, dy2, dy3, dy4,
  u1, u2, u3, u4, du1, du2, du3, du4,
  v1, v2, v3, v4, dv1, dv2, dv3, dv4;

/*This constant is used to get the maximum offset into the texture. Since
we use 64x64 textures, the maximum offset is 63, which is shifted left by 16
bits to gain accuracy. Note: We do this for all our coordinates!!*/
  long const Max = 63L << 16;

  //Assume extremes
  miny = MAXLONG;
  maxy = -MAXLONG;

  //Find minimum and maximum y coordinates
  if (y1 < miny) miny = y1; if (y1 > maxy) maxy = y1;
  if (y2 < miny) miny = y2; if (y2 > maxy) maxy = y2;
  if (y3 < miny) miny = y3; if (y3 > maxy) maxy = y3;
  if (y4 < miny) miny = y4; if (y4 > maxy) maxy = y4;

  //Calculate delta y and delta x for first side
  dy1 = y2 - y1; dx1 = x2 - x1;
  //If the line is not horizontal, a slope can be calculated
  if (dy1 != 0) {
    //Calculate how the x coordinate should change when moving one pixel down
    dx1 = (dx1 << 16) / dy1;
    //Calculate how the texture coordinates should change
    /*NOTE: Moving to the right in the bitmap, from vertex 1 to 2
      (du is positive, dv is zero)*/
    du1 = (64L << 16) / dy1;
    dv1 = 0;
  }

  //Do the same for the other three sides
  dy2 = y3 - y2; dx2 = x3 - x2;
  if (dy2 != 0) {
    dx2 = (dx2 << 16) / dy2;
    /*NOTE: Moving downwards in the bitmap, from vertex 2 to 3
      (du is zero, dv is positive)*/
    du2 = 0;
    dv2 = (64L << 16) / dy2;
  }

  dy3 = y4 - y3; dx3 = x4 - x3;
  if (dy3 != 0)  {
    dx3 = (dx3 << 16) / dy3;
    /*NOTE: Moving to the left in the bitmap, from vertex 3 to 4
      (du is negative, dv is zero)*/
    du3 = -(64L << 16) / dy3;
    dv3 = 0;
  }

  dy4 = y1 - y4; dx4 = x1 - x4;
  if (dy4 != 0)  {
    dx4 = (dx4 << 16) / dy4;
    /*NOTE: Moving upwards in the bitmap, from vertex 2 to 3
      (du is zero, dv is negative)*/
    du4 = 0;
    dv4 = -(64L << 16) / dy4;
  }

  //Check to see whether vertex 2 is the topmost vertex of the first side
  if (y2 < y1)  {
    //If so, start at the coordinates of vertex 2
    u1 = Max; v1 = 0; tx1 = x2;
  } else {
    //Otherwise, start at vertex 1
    u1 = 0; v1 = 0;  tx1 = x1;
  }

  //Do the same for the other sides
  if (y3 < y2) {
    u2 = Max; v2 = Max; tx2 = x3;
  } else {
    u2 = Max; v2 = 0;  tx2 = x2;
  }

  if (y4 < y3) {
    u3 = 0; v3 = Max; tx3 = x4;
  } else {
    u3 = Max; v3 = Max; tx3 = x3;
  }

  if (y1 < y4) {
    u4 = 0; v4 = 0; tx4 = x1;
  } else {
    u4 = 0; v4 = Max; tx4 = x4;
  }

  //Now shift the coordinates left 16 bits to gain accuracy
  tx1 <<= 16;
  tx2 <<= 16;
  tx3 <<= 16;
  tx4 <<= 16;

  /*Here we start interpolation. The following subroutine processes all the
  scan lines in the polygon*/
  for (y = miny; y <= maxy; y++) {

    //Assume extremes
    minx = MAXLONG;
    maxx = -MAXLONG;

    //Check whether y is a valid value for the first side
    if ((y >= y1) || (y >= y2)) {
      if ((y <= y1) || (y <= y2)) {
        if (tx1 < minx) {
          /*If necessary, adjust the start point of the scan line and store
          appropriate texture coordinates.*/
          minx = tx1;
          minu = u1;
          minv = v1;
        }
        if (tx1 > maxx) {
          /*If necessary, adjust the end point of the scan line and store
          appropriate texture coordinates.*/
          maxx = tx1;
          maxu = u1;
          maxv = v1;
        }
        //Move on to the next pixel on this side
        tx1 += dx1;
        //And its texture coordinates
        u1 += du1;
        v1 += dv1;
      }
    }

    //Process the other lines in the same manner
    if ((y >= y2) || (y >= y3)) {
      if ((y <= y2) || (y <= y3)) {
        if (tx2 < minx) {
          minx = tx2;
          minu = u2;
          minv = v2;
        }
        if (tx2 > maxx) {
          maxx = tx2;
          maxu = u2;
          maxv = v2;
        }
        tx2 += dx2;
        u2 += du2;
        v2 += dv2;
      }
    }

    if ((y >= y3) || (y >= y4)) {
      if ((y <= y3) || (y <= y4)) {
        if (tx3 < minx) {
          minx = tx3;
          minu = u3;
          minv = v3;
        }
        if (tx3 > maxx) {
          maxx = tx3;
          maxu = u3;
          maxv = v3;
        }
        tx3 += dx3;
        u3 += du3;
        v3 += dv3;
      }
    }

    if ((y >= y4) || (y >= y1)) {
      if ((y <= y4) || (y <= y1)) {
        if (tx4 < minx) {
          minx = tx4;
          minu = u4;
          minv = v4;
        }
        if (tx4 > maxx) {
          maxx = tx4;
          maxu = u4;
          maxv = v4;
        }
        tx4 += dx4;
        u4 += du4;
        v4 += dv4;
      }
    }

    /*Now that we have the start and endpoints, Their extra precision should
    be removed. It is no longer needed. The texture coordinates, however
    shouldn't yet be changed since the "DrawTLine" function still has to
    interpolate between them*/
    minx >>= 16;
    maxx >>= 16;
    //if ((minu | maxu | minv | maxv) == 0) {printf ("%c", 7);}
    //And here we draw the line
    DrawTLine (minx, maxx, y, minu, minv, maxu, maxv, Sprite);
  }
}

void main ()
{
  //This is used to store a texture. We use 64x64 textures for this demo
  char *Texture;
  unsigned char p, q, c;

  //Switch to VGA 320x200 256 colour
  VGAScreen ();
  #ifdef __DJGPP
     Screen = (char *) (__djgpp_conventional_base + 0xA0000);
    __djgpp_nearptr_enable ();
  #endif
  //Create a new texture
  Texture = (char *) malloc (4096);
  /*Fill it with a pattern and display the same pattern on the top left
  of the screen so you can see how it is mapped*/
  for (p = 0; p < 64; p++)
    for (q = 0; q < 64; q++) {
      c = ((q >> 2) << 4) + (p >> 2);
      Texture [(q << 6) + p] = c;
      PutPixel (p, q, c);
    }
  //Display it on the screen mapped onto a polygon
  TextureMap (120, 20, 280, 50, 250, 170, 20, 140, Texture);
  //Wait for a key press
  getch ();
  //Switch back to text mode
  TextScreen ();
}